local COOKIE_MAX_AGE = 360

local json = require("json")
local utils = require("utils")
local const = require("const")
local Set = require("set")
local resty_string = require("resty.string")
local hmac = require("resty.hmac")
local resty_random = require("resty.random")
local pl_pretty = require("pl.pretty")


local memkv = ngx.shared.memkv
local request_method = ngx.var.request_method

utils.check_login()
_, login_token = utils.check_permission("GatewayManage")

-- get user name
local user_name = memkv:get(login_token .. "." .. "username")

local function set_gateway_3tuple(pk, dn, ds)
    local params = {}
    table.insert(params, {pk, dn, ds})
    local ok, msg = utils.set_data(const.data_key.dimu.device_3tuple, json.encode(params))
    if (ok) then
        return true
    end

    utils.log_error("set gateway 3tuple failed" .. msg)
    utils.error_exit(500, "set gateway 3tuple failed")
end


if "POST" == request_method then
    --[[
        "User": string["admin"],
        "Password": string,
        "ProductKey": string,
        "DeviceName": string,
        "DeviceSecret": string,
        "DeviceAuthMode": string["Auto", "Manual"], //配置子设备接入方式；Auto：无需认证，自动接入；Manual：自动接入的设备需在控制台上确认
        "ChannelQos": int[0, 1, 2] //可选参数，暂时不用
    --]]
    ngx.req.read_body()
    local args_json = ngx.req.get_body_data()
    args = json.decode(args_json)
    if (nil == args) then
        utils.error_exit(400)
    end

    local user = args["User"]
    if (user ~= user_name) then
        utils.error_exit(400)
    end

    local passwd = args["Password"]
    local pk = args["ProductKey"]
    local dn = args["DeviceName"]
    local ds = args["DeviceSecret"]
    local auth_mode = args["DeviceAuthMode"]

    utils.check_username(user)
    utils.check_passwd(passwd)
    utils.check_pk(pk)
    utils.check_dn(dn)
    utils.check_ds(ds)
    utils.check_authmode(auth_mode)

    if ("admin" ~= user) then
        utils.error_exit(400, "invalid user name")
    end

    local user_list = {}
    ok, msg = utils.get_data(const.data_key.eweb.user)
    if (ok) then
        if (msg) then
            user_list = json.decode(msg)
        else
            user_list["UserList"] = {}
        end
    else
        utils.log_error("get user data failed")
        utils.error_exit(500)
    end

    local user_info = {}

    if ((nil == next(user_list["UserList"])) or (passwd ~= user_list["UserList"][user]["PasswordHash"])) then
        if (string.len(passwd) > 20) then
            utils.log_warn(ngx.INFO, "invalid format of the passwd")
            utils.error_exit(400, "invalid parameter")
        end

        salt = resty_string.to_hex(resty_random.bytes(16))
        user_info["PasswordSalt"] = salt
        hex = utils.hmac_sha256(salt, passwd)
        if (nil == hex) then
            utils.log_error("hmac failed")
            utils.error_exit(500)
        end
        user_info["PasswordHash"] = resty_string.to_hex(hex)
        utils.log_info("admin's password be changed")
    else 
        user_info["PasswordSalt"] = user_list["UserList"][user]["PasswordSalt"]
        user_info["PasswordHash"] = user_list["UserList"][user]["PasswordHash"]
        utils.log_info("admin's password unchanged")
    end

    user_info["Username"] = user
    -- TODO hide ComputeManage module
    --user_info["Permission"] = {"GatewayManage", "DeviceManage", "DriverManage", "ComputeManage", "LogManage", "UserManage"}
    user_info["Permission"] = {"MessageRoute", "TSLManage", "StreamCompute","GatewayManage", "DeviceManage", "DriverManage", "LogManage", "UserManage"}
    user_info["LastLoginTime"] = tostring(os.time())

    user_list["UserList"][user] = user_info
    user_list_raw = json.encode(user_list)
    utils.log_info("user list:", user_list_raw)

    -- Delete .sst/ and /var/tmp/iiot_aliyun
    utils.call_shell('./script/stop_gw.sh')
    utils.call_shell('./script/reset_gw.sh')
    utils.call_shell('./script/start_gw.sh')

    ok, msg = utils.set_data(const.data_key.eweb.user, user_list_raw)
    if (not ok) then
        utils.log_error("store user data failed" .. msg)
        utils.error_exit("store user data failed")
    end

    -- Write gateway configuration
    -- config: ProductKey, DeviceName, DeviceSecret, DeviceAuthMode, DeviceTags
    utils.set_gw_config(pk, dn, ds, auth_mode)

    -- import device 3tuple into dimu
    set_gateway_3tuple(pk, dn, ds)

    -- update the user's permission
    local permission = ""
    for _, val in ipairs(user_info["Permission"]) do 
        permission = val .. "," .. permission
    end
    memkv:set(login_token, permission, COOKIE_MAX_AGE)

    -- update product key list
    --local pk_set = Set:new({})
    --Set.add(pk_set, pk)
    --utils.set_product_key_list( Set.tolist(pk_set))
    -- utils.set_product_key_list("[\""..pk.."\"]")
    
    local file = io.open("../../config/gateway-monitor/product-config.json", 'r') 
    local gw_tsl = file:read("*a")
    file:close()
   
    if gw_tsl ~= nil then
        utils.set_data(const.data_key.tsl.standard_tsl_prefix..pk, gw_tsl)
    end

    -- call shell script to restart gateway system except eweb
    ngx.sleep(1.0)
    local ok, retcode, _, _ = utils.call_shell('./script/start_gw.sh')
    
    local params = {}
    params["params"] = {}
    
    local item = {}
    table.insert(item, pk)
    table.insert(item, dn)
    table.insert(params["params"], item)

    utils.dbus_call_edge(const.dimu.wkn,
                            const.dimu.obj,
                            const.dimu.inf,
                            "authorizeDevice",
                            json.encode(params))

    if (ok and 0 == retcode) then
        utils.error_exit(200)
    else
        utils.error_exit(500, 'restart gateway failed')
    end

elseif "GET" == request_method then
    local ok, msg_user = utils.get_data(const.data_key.eweb.user)
    if (not ok) then
        utils.log_error("get user data failed" .. msg_user .. msg_gwconfig)
        utils.error_exit(500, "get data fail")
    end

    if (nil == msg_user) then 
        utils.error_exit(400, "gateway have not been initialized")
    end

    local user_data = json.decode(msg_user)
    local user_info = user_data["UserList"]["admin"]
    if (nil == user_info) then
        utils.error_exit(500, "user info is incorrect.")
    end

    local res = {}
    res["User"] = "admin"
    res["PasswordHash"] = user_info["PasswordHash"]
    res["LastLoginTime"] = user_info["LastLoginTime"]

    local pk, dn, ds, auth_mode = utils.get_gw_config()
    res["ProductKey"] = pk
    res["DeviceName"] = dn
    res["DeviceSecret"] = ds
    res["DeviceAuthMode"] = auth_mode
    res["DeviceTags"] = utils.get_device_tags(pk, dn)
    res_raw = json.encode(res)
    utils.error_exit(200, res_raw)
end

utils.error_exit(400, "unsupported request method")

------ EOF ------
