local const = require("const")
local json = require("json")
local resty_string = require("resty.string")
local hmac = require("resty.hmac")
local resty_random = require("resty.random")
local pl_utils = require("pl.utils")
local Set = require("set")
local ldbus = require "ldbus"
local utils = { _version = "0.1.0" }

local logger = require("lua_logger")

function split(str, sep)
    local t = {}
    string.gsub(str, '[^'..sep..']+', function(w) table.insert(t, w) end )
    return t
end

function utils.split_str(str, sep)
   return split(str, sep)
end

function utils.hmac_sha256(key, data)
    local hmacsha256 = hmac:new(key, hmac.ALGOS.SHA256)
    if not hmac then
        return nil
    end

    local ok = hmacsha256:update(data)
    if not ok then
        return nil
    end

    local hmac_binary = hmacsha256:final()
    return hmac_binary
end

local function in_array(array, item)
    if not array then
        return false
    end

    if array then
        for k, v in ipairs(array) do
            if v == item then
                return true
            end
        end
    end

    return false
end


function utils.is_in_array(array, item)
    return in_array(array, item)
end


function utils.check_login()
    local ck = require("resty.cookie")
    local cookie, err = ck:new()
    if not cookie then
        utils.error_exit(401)
    end

    print("client ip is: " .. ngx.var.remote_addr)

    -- get login cookie
    local login_token, err = cookie:get("loginToken")
    if login_token then
        local memkv = ngx.shared.memkv
        perm = memkv:get(login_token)
        if (perm) then
            return true, login_token, perm
        end
    end

    utils.error_exit(401)
end


function utils.check_permission(...)
    local is_login, token, perm = utils.check_login()
    if (is_login) then
        --local ngx_re = require "ngx.re"
        --local perm_list, err = ngx_re.split(perm, ",")
        local perm_list = split(perm, ",")
        if (perm_list) then
            for _, item in ipairs{...} do
                if (in_array(perm_list, item)) then
                    return true, token
                end
            end
        end
    end
    print("----->>>>>>>  "..perm)
    utils.error_exit(403)
end


function utils.error_exit(err, msg)
    if err == 200 or err == 460 then
        if msg then
	    res = '{"code": 200, "message": "success", "data":' .. msg .. '}'
            ngx.say(res)
        else
            ngx.say([[{"code": 200, "message": "success", "data": {}}]])
        end
    elseif err == 400 then
        if msg then
            res = '{"code": 400, "message": "Bad Request: ' .. msg .. '", "data": {}}'
            ngx.say(res)
        else -- if msg then
            ngx.say([[{"code": 400, "message": "Bad Request", "data": {}}]])
        end
    elseif err == 401 then
        if msg then
            res = '{"code": 401, "message": "Unauthorized: ' .. msg .. '", "data": {}}'
            ngx.say(res)
        else -- if msg then
            ngx.say([[{"code": 401, "message": "Unauthorized", "data": {}}]])
        end

    elseif err == 401.1 then
        if msg then
            res = '{"code": 401.1, "message": "Invalid username or password:' .. msg .. '", "data": {}}'
            ngx.say(res)
        else -- if msg then
            ngx.say([[{"code": 401.1, "message": "Invalid username or password", "data": {}}]])
        end
    elseif err == 403 then
        if msg then
            res = '{"code": 403, "message": "Forbidden: ' .. msg .. '", "data": {}}'
            ngx.say(res)
        else -- if msg then
            ngx.say([[{"code": 403, "message": "Forbidden", "data": {}}]])
        end
    elseif err == 404 then
        if msg then
            res = '{"code": 404, "message": "Not Found: ' .. msg .. '", "data": {}}'
            ngx.say(res)
        else -- if msg then
            ngx.say([[{"code": 404, "message": "Not Found", "data": {}}]])
        end
    elseif err == 500 then
        if msg then
	    res = {}
	    res["code"] = err 
	    res["data"] = ""
	    res["message"] = msg
            ngx.say(json.encode(res))
        else -- if msg then
            ngx.say([[{"code": 500, "message": "Internal Service Error", "data": {}}]])
        end
    elseif err == 503 then
        if msg then
            res = '{"code": 503, "message": "Service Unavalible: ' .. msg .. '", "data": {}}'
            ngx.say(res)
        else -- if msg then
            ngx.say([[{"code": 503, "message": "Service Unavalible", "data": {}}]])
        end
    else
        ngx.say([[{"code": 500, "message": "Internal Service Error", "data": {}}]])
    end
    ngx.exit(ngx.HTTP_OK)
end


function utils.get_item_index(total, page_size, current_page)
    if (total <= 0 or page_size <= 0 or current_page <= 0) then
        return 0, -1
    end

    local min = page_size * (current_page - 1) + 1
    local max = page_size * current_page
    if (total < max) then
        max = total
    end

    return min, max
end


function utils.call_shell(cmd)
    local ok, retcode, stdout, stderr = pl_utils.executeex(cmd)
    return ok, retcode, stdout, stderr
end

local function dbus_send(bus, wkn, obj, inf, method, ...)
    local conn = ldbus.connection.open(bus)
    if not (conn)
    then
        utils.log_error(">>>> failed to open dbus")
    end

    if not (ldbus.bus.register(conn, "register failed"))
    then
        utils.log_error(">>>>failed to register dbus")
    end
    
    local msg = ldbus.message.new_method_call (wkn, obj, inf, method)
    if not (msg)
    then
        utils.log_error(">>>>>method call faild")
    end 
    
    local iter = ldbus.message.iter.new()
    msg:iter_init_append(iter)
   
    local params = {}
    for _, arg in ipairs{...} do
        assert( iter:append_basic (arg), "out of memory")
        table.insert(params, arg)
    end
    
    local ret = {}
    --print("get dbus param: bus" .. bus .. "  wkn: " .. wkn .. "  obj: " .. obj .. "  inf: " .. inf .. "  method: ".. method.."  params:  "..json.encode(params))
    local timeout = nil
    if (wkn == const.function_compute.wkn)
    then
        timeout = 1800
    end

    local reply, err_msg = conn:send_with_reply_and_block (msg, timeout)
    if not reply
    then
        utils.log_error("failed to send method call:  "..type(reply).."  type msg: "..err_msg)
        table.insert(ret, err_msg)
        conn:close()
        return "ok", ret, ""
    end
    assert( reply:iter_init( iter ), "reply message has no arguments")
  
    table.insert(ret, iter:get_basic())
    if (iter:get_arg_type() == ldbus.types.int32 and iter:next())
    then
        assert(iter:get_arg_type() == ldbus.types.string, "argument not string.")
        table.insert(ret, iter:get_basic())
    end
    ok = "ok"
    conn:close()
    --    utils.log_debug("dbus return: "..'"'..json.encode(ret)..'"')
    return ok, ret, "stderr"
end


function utils.dbus_call_edge(wkn, obj, inf, method, ...)
    return dbus_send("unix:path=/tmp/var/run/mbusd/mbusd_socket", wkn, obj, inf, method, ...)
end


function utils.set_data(key, val)
    local ok, stdout, stderr = utils.dbus_call_edge("iot.dmp.configmanager",
                            "/iot/dmp/configmanager",
                            "iot.dmp.configmanager",
                            "set_config",
                            key,
                            val)

    if (ok) then
        int32_code = tonumber(stdout[1]) 
        if (0 == int32_code) then
            return true, "success"
        end
    end

    return false, stdout[2] .. stderr
end

function utils.get_core_connection_state()
    local ok, stdout, err = utils.dbus_call_edge(
            "iot.bridge.mqtt",
            "/iot/bridge/mqtt",
            "iot.bridge.mqtt",
            "getGatewayNetState")

    if not ok then
        return nil, err
    end

    local state = {}
    if stdout[1] == "connected" then
        state["connected"] = true
        ok, stdout, err = utils.dbus_call_edge(
                "iot.bridge.mqtt",
                "/iot/bridge/mqtt",
                "iot.bridge.mqtt",
                "getOnlineTimestampLast")
        repeat
            if not ok and type(stdout[1] ~= "string") then break end

            local res
            ok, res = pcall(json.decode, stdout[1])
            if not ok then break end

            if res["code"] ~= 0 then
                utils.log_info("get online timestamp: %s", res["message"])
                break
            end

            state["timestamp"] = res["payload"]["time"]
        until true
    else
        state["connected"] = false
    end

    return state, nil
end

function utils.get_core_host_state()
    local ok, ret, err = utils.dbus_call_edge(
            "iot.dmp.service.monitor",
            "/iot/dmp/service/monitor",
            "iot.dmp.service.monitor",
            "QueryHostInfo",
            "[]")

    if not ok then return nil, err end

    local res
    ok, res = pcall(json.decode, ret[1])
    if not ok then return nil, "not a JSON" end

    return res, nil
end

function utils.get_core_host_properties()
    local state, err = utils.get_core_host_state()
    if err then
        return nil, err
    end

    local properties = {
        gateway_ip = "172.20.0.1",
    }
    if state.cpu then
        properties.cpu_core_number = state.cpu.core_number
        properties.cpu_usage = state.cpu.usage
    end
    if state.memory then
        properties.memory_total = state.memory.total
        properties.memory_free = state.memory.free
        properties.memory_usage = state.memory.usage
    end
    if state.disk then
        properties.disk_usage = state.disk.usage
    end
    if state.system then
        properties.system_info = state.system.info
    end
    return properties, nil
end

function utils.get_core_version()
    local file = io.open("/linkedge/gateway/build/VERSION", "r")
    local version = file:read()
    file:close()
    return version
end

function utils.get_standard_tsl_key_list()
    local ok, stdout, stderr = utils.dbus_call_edge("iot.dmp.configmanager",
                                "/iot/dmp/configmanager",
                                "iot.dmp.configmanager",
                                "get_config_keylist",
                                const.data_key.tsl.standard_tsl_prefix.."*")
    if (ok) then
        int32_code = tonumber(stdout[1])
        if (0 == int32_code) then
            msg = stdout[2]
            local ret = {}
            for k, v in pairs(json.decode(msg)) do
                if string.sub(v, 1, #const.data_key.tsl.extend_tsl_prefix) ~= const.data_key.tsl.extend_tsl_prefix then
                    table.insert(ret, v)
                end
            end

            if ("" ~= msg) then
                return json.encode(ret)
            end
        end
        -- ngx.log(ngx.WARN, "int code:"..int32_code.."value of the key:", key, " is empty. ", stdout[2])
        return "{}"
    end

    utils.log_error("get data failed" .. stdout[2] .. stderr)
    utils.error_exit(500, "get data failed")
end

function utils.get_extend_tsl_key_list()
    local ok, stdout, stderr = utils.dbus_call_edge("iot.dmp.configmanager",
                                "/iot/dmp/configmanager",
                                "iot.dmp.configmanager",
                                "get_config_keylist",
                                const.data_key.tsl.extend_tsl_prefix.."*")
    if (ok) then
        int32_code = tonumber(stdout[1])
        if (0 == int32_code) then
            msg = stdout[2]
            if ("" ~= msg) then
                return msg
            end
        end
        -- ngx.log(ngx.WARN, "int code:"..int32_code.."value of the key:", key, " is empty. ", stdout[2])
        return "{}"
    end

    utils.log_error("get data failed" .. stdout[2] .. stderr)
    utils.error_exit(500, "get data failed")
end

function utils.get_data(key)
    local ok, stdout, stderr = utils.dbus_call_edge("iot.dmp.configmanager",
                                "/iot/dmp/configmanager",
                                "iot.dmp.configmanager",
                                "get_config",
                                key)
    if (ok) then
        int32_code = tonumber(stdout[1])
        if (0 == int32_code) then
            -- strip the spaces in the header
            msg = stdout[2] 
            if ("" ~= msg) then
                return true, msg
            end
        end
        -- ngx.log(ngx.WARN, "int code:"..int32_code.."value of the key:", key, " is empty. ", stdout[2])
        return true, nil
    end

    utils.log_error("get data failed" .. stdout[2] .. stderr)
    utils.error_exit(500, "get data failed")
end


local function is_match(orign, rule, len_min, len_max)
    if (nil == orign) then
        return false
    end

    if (string.match(orign, rule)) then
        len = string.len(orign)
        if (len >= len_min and len <= len_max) then
            return true
        end
    end

    return false
end


local function check_invalid_exit(orign, rule, len_min, len_max, msg)
    if (not is_match(orign, rule, len_min, len_max)) then
        utils.error_exit(400, msg)
    end

    return true
end


function utils.check_username(username)
    if (nil == username or type(username) ~= "string" or #username == 0) then
        utils.error_exit(400, "username type error: should be string.")
    end
    return check_invalid_exit(username, "^%w+$", 5, 20, "invalid username")
end


function utils.check_passwd(passwd)
    if (nil == passwd or type(passwd) ~= "string" or #passwd == 0) then
        utils.error_exit(400, "password type error: should be string.")
    end

    if (64 == string.len(passwd)) then --password, 64 chars
        return true
    end
    return check_invalid_exit(passwd, "^%w+$", 5, 20, "invalid password")
end


function utils.check_pk(pk)
    return check_invalid_exit(pk, "^%w+$", 11, 11, "invalid product key")
end


function utils.check_dn(dn)
    -- 支持英文字母、数字和特殊字符-_@.:，长度限制4~32
    return check_invalid_exit(dn, "^[%w%_%-%.%@%:]+$", 4, 32, "invalid device name")
end


function utils.check_ds(ds)
    return check_invalid_exit(ds, "^%w+$", 32, 32, "invalid device secret")
end


function utils.check_authmode(authmode)
    if ("Auto" == authmode or "Manual" == authmode) then
        return true
    end

    utils.error_exit(400, "invalid authmode")
end


function utils.check_tag(tag)
    -- empty or table list
    type_tag = type(tag)
    if (nil == tag or "string" == type_tag or "table" == type_tag) then
        return true
    end

    utils.error_exit(400, "invalid tag")
end


function utils.check_permission_format(permission)
    valid = {"GatewayManage", "DeviceManage", "DriverManage", "LogManage", "UserManage", "MessageRoute", "TSLManage", "StreamCompute", "FunctionComputeManage"}

    for _, val in ipairs(permission) do
        if (not in_array(valid, val)) then
            utils.error_exit(400, "invalid permission format")
        end
    end

    return true
end


function utils.check_current_page( cur_page )
    if (cur_page and "number" == type(cur_page) and cur_page > 0) then
        return true
    end

    utils.error_exit(400, "invalid current page")
end

function utils.check_page_size_allow_zero( page_size )
    
    if page_size == nil then
        return 0
    end

    if type(page_size) == "number" and page_size >= 0 then
        return  page_size
    end

    if type(page_size) == "string" and #page_size == 0 then
        return 0
    end

    utils.error_exit(400, "invalid page size")
end


function utils.check_page_size( page_size )
    if (page_size and "number" == type(page_size) and page_size > 0) then
        return true
    end

    utils.error_exit(400, "invalid page size")
end


-- get caller infomation from debug.traceback
-- output like: /src/utils.lua:370: in function 'log_info':
local function get_log_prefix()

    --[[ sample of traceback
    stack traceback:
	    ./src/utils.lua:350: in function 'get_log_prefix'
	    ./src/utils.lua:370: in function 'log_info'
	    /var/nginx/eweb/src/login.lua:52: in function </var/nginx/eweb/src/login.lua:1>, client: 30.6.57.149,
    ]]--

    tb = debug.traceback()
    tb_list = split(tb, '\n')
    assert(tb_list[4], "tb_list[4] should never be nil")

    log_prefix = string.gsub(tb_list[4], "^(%s*)(.+)$", "%2")
    -- log_prefix = string.gsub(tb_list[4], "^(%s+)(.+) in function '(%w+)'", "%2%3(): ")
    -- if (nil == log_prefix) then
    --     log_prefix = string.gsub(tb[4], "^(%s+)(.+) in function <(.+)>", "%2%3(): ")
    -- end
    --return log_prefix

    return log_prefix
end

local function check_init_log()
    local memkv = ngx.shared.memkv
    local flag = memkv:get("flag_log_inited") 
    
    if not flag then
        logger.log_init("web_console", 1, 0, 2)
        ngx.ctx.init_logger = 1
        memkv:set("flag_log_inited", 1)
        ngx.log(ngx.ERR, "init logger")       
    end
end

local function log_dump(lvl, msg)
    local g_logger_switch = 1
    if (g_logger_switch == 1) then
        local func_dispatch = nil
        check_init_log()
        if (lvl == ngx.DEBUG) then
            func_dispatch = logger.log_d
        elseif (lvl == ngx.INFO) then
            func_dispatch = logger.log_i
        elseif (lvl == ngx.ERR) then
            func_dispatch = logger.log_e
        elseif (lvl == ngx.CRIT) then
            func_dispatch = logger.log_m
        elseif (lvl == ngx.WARN) then
            func_dispatch = logger.log_w
        elseif (lvl == ngx.INFO) then
            func_dispatch = logger.log_i
        end
        func_dispatch("local_console", msg)
    else
        ngx.log(lvl, msg)
    end
end

function utils.log_debug(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)

    log_dump(ngx.DEBUG, msg)
end

function utils.log_info(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)
    log_dump(ngx.INFO, msg)
end

function utils.log_warn(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)
    log_dump(ngx.WARN, msg)
end

function utils.log_error(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)
    log_dump(ngx.ERR, msg)
end

function utils.log_critical(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)
    log_dump(ngx.CRIT, msg)
end


function utils.set_gw_config(pk, dn, ds, auth_mode)
    local ok1, msg1 = utils.set_data(const.data_key.gwconfig.productkey, pk)
    local ok2, msg2 = utils.set_data(const.data_key.gwconfig.devicename, dn)
    local ok3, msg3 = utils.set_data(const.data_key.gwconfig.devicesecret, ds)
    if (not (ok1 and ok2 and ok3)) then
        utils.log_error("set gateway config failed" .. msg1 .. msg2 .. msg3)
        utils.error_exit(500, "set data failed")
    end

    local auto_auth = "off" -- in off mode by default
    if ("Auto" == auth_mode) then
        auto_auth = "on"
    end

    utils.set_data(const.data_key.dimu.auto_auth_sw, auto_auth)
    if (not ok) then
        utils.log_error("set gateway auto auth value failed" .. msg)
        utils.error_exit(500, "set data failed")
    end
end


-- return: pk, dn, secret, authmode, tags
function utils.get_gw_config( )
    -- get gateway 3tuple
    local ok1, msg_pk = utils.get_data(const.data_key.gwconfig.productkey)
    local ok2, msg_dn = utils.get_data(const.data_key.gwconfig.devicename)
    local ok3, msg_ds = utils.get_data(const.data_key.gwconfig.devicesecret)
    if (not (ok1 and ok2 and ok3)) then
        utils.log_error("get gateway config failed" .. msg_pk .. msg_dn .. msg_ds)
        utils.error_exit(500, "get data failed")
    end

    if (nil == msg_pk or nil == msg_dn or nil == msg_ds) then
        return nil, nil, nil, nil
    end

    -- get auth mode
    local auth_mode = "Manual" -- 'Manual' mode by default
    local auto_auth = utils.get_data(const.data_key.dimu.auto_auth_sw)
    if (not ok) then
        utils.log_error("get gateway auto auth value failed" .. msg)
        utils.error_exit(500, "get data failed")
    end
    if (auto_auth and "on" == auto_auth) then
        auth_mode = "Auto"
    end

    return msg_pk, msg_dn, msg_ds, auth_mode
end


function utils.set_driver_config(driver, config)
    local key = const.data_key.driver.prefix .. driver
    local ok, msg = utils.set_data(key, config)
    if (not ok) then
        utils.log_error("store driver config failed" .. msg)
        utils.error_exit(500, "store config failed")
    end
end

function utils.get_driver_config(driver) 
    local key = const.data_key.driver.prefix .. driver
    local ok, msg = utils.get_data(key)
    if (not ok) then
        -- utils.error_exit(500, "get config failed")
        return nil
    end

    return msg
end

function utils.set_device_tags(pk, dn, tags)
    local cjson = require('cjson')
    local device_tags = {}
    local params = {}
    params["params"] = {}
    params["params"]["devices"] = {pk .. '_' .. dn}
    params["params"]["tags"] = {}
    for _, item in ipairs(tags) do
        local res = utils.split_str(item, '=')
        params["params"]["tags"][res[1]] = res[2]
    end

    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                    const.dimu.obj,
                                    const.dimu.inf,
                                    "setDeviceTags",
                                    cjson.encode(params))

    if (ok == "ok") then
        local ok, msg = pcall(json.decode, msg_raw[1])
        if ok then
            if 0 == msg["code"] then
                utils.error_exit(200, json.encode(""))
            else
                local ret = {}
                local ret_error = {}
                if not (msg["params"]["failed_message"] and type(msg["params"]["failed_message"]) == "table" )then
                 utils.error_exit(400, "failed to set device tags.")
                end
                for k, v in pairs(msg["params"]["failed_message"]) do
                    if v["code"] ~= 0 then
                        local item = {}
                        local tmp = utils.split_str(v["params"]["device"], "_")
                        item["ProductKey"] = tmp[1]
                        item["DeviceName"] = table.concat(tmp, "_", 2)
                        item["FaileMessage"] = v["message"]
                        table.insert(ret_error, item)
                    end
                end
                ret["code"] = 400
                ret["data"] = ret_error
                ret["message"] = ""
                ngx.say(json.encode(ret))
                ngx.exit(ngx.HTTP_OK)
            end
        end
        if (ok and 0 == msg["code"]) then
            utils.error_exit(200, json.encode(""))
        end
    else
        utils.log_error("setDeviceTags failed" .. msg)
        utils.error_exit(500, "store device tags failed")
    end
end

function utils.get_device_tags(pk, dn)
    local device_tags = {}
    
    if(pk == nil or dn == nil) then
        return device_tags
    end

    local cloudid = pk .. '_' .. dn
    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                    const.dimu.obj,
                                    const.dimu.inf,
                                    "getDeviceTags",
                                    '{"params": ["' .. cloudid .. '"]}')
    if (ok and msg_raw) then
        local ok, msg = pcall(json.decode, msg_raw[1])
        if (ok and 0 == msg["code"]) then
            local tags = msg["params"][cloudid]
            for key, val in pairs(tags) do
                table.insert(device_tags, key .. "=" .. val)
            end
        end
    else
        utils.log_error("get nothing from getDeviceTags:" .. err)
    end

    return device_tags
end

function utils.sync_device_tags()
    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                    const.dimu.obj,
                                    const.dimu.inf,
                                    "syncTag")

    if (ok) and (tonumber(msg_raw[1]) == 0) then
        return true, "ok"
    else
        utils.log_error("syncDeviceTags failed : " .. msg_raw[1])
        return false, msg_raw[1]
    end
end

function utils.set_product_key_list( product_keys )
    local ok, msg = utils.set_data(const.data_key.eweb.product_key_list, json.encode(product_keys))
    if (ok) then
        return true
    end

    utils.log_error("set product key list failed" .. msg)
    utils.error_exit(500, "store product key list failed")
end


function utils.get_product_key_list( )
    local ok, msg = utils.get_data(const.data_key.eweb.product_key_list)
    if (ok and msg) then
        local ok, list = pcall(json.decode, msg) 
        if (ok) then
            return list
        end
    else 
        return {}
    end

    utils.log_error("get device tuple data failed" .. msg)
    utils.error_exit(500, "get device tuple failed")
end


-- get product key and device name by product key and localid
-- function utils.pklid2pkdn(pk, lid)
--     local params = {}
--     params["params"] = {}
--     table.insert(params["params"], {pk, lid})

--      local ok, stdout, stderr = utils.dbus_call_edge(
--                                     const.dimu.wkn,
--                                     const.dimu.obj,
--                                     const.dimu.inf,
--                                     "getLocalDeviceState", 
--                                     json.encode(params))  
--     if (ok) then 
--         local ok, ret = pcall(json.decode, stdout)
--         if ok then 
--             return true, pk, ret["params"]["devices"][1]["deviceName"]
--         end
--     end
    
--     return false, nil, nil
-- end

return utils

------ EOF ------
