local const = require("const")
local json = require("json")
local hmac = require("resty.hmac")
local pl_utils = require("pl.utils")
local ldbus = require("ldbus")
local utils = { _version = "0.1.0" }

local logger = require("lua_logger")

function split(str, sep)
    local t = {}
    string.gsub(str, '[^'..sep..']+', function(w) table.insert(t, w) end )
    return t
end

function utils.is_array_table(t)
    if type(t) ~= "table" then
        return false
    end

    local n = #t
    for i, _ in pairs(t) do
        if type(i) ~= "number" then
            return false
        end

        if i > n then
            return false
        end
    end

    return true
end

function utils.split_str(str, sep)
   return split(str, sep)
end

function utils.hmac_sha256(key, data)
    local sha256 = hmac:new(key, hmac.ALGOS.SHA256)
    if not sha256 then return nil end

    local ok = sha256:update(data)
    if not ok then return nil end

    return sha256:final()
end

local function in_array(array, item)
    if not array then
        return false
    end

    if array then
        for k, v in ipairs(array) do
            if v == item then
                return true
            end
        end
    end

    return false
end


function utils.is_in_array(array, item)
    return in_array(array, item)
end

function utils.check_auth()
    local resty_cookie = require("resty.cookie")
    local cookie, err = resty_cookie:new()
    if not cookie then
        utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, err)
    end

    print("client ip is: " .. ngx.var.remote_addr)

    -- get access cookie
    local token, err = cookie:get("token")
    if not token then
        utils.error_exit(ngx.HTTP_UNAUTHORIZED, "unauthorized")
    end

    local memkv = ngx.shared.memkv
    local perm = memkv:get(token)
    if not perm then
        utils.error_exit(ngx.HTTP_UNAUTHORIZED, "unauthorized")
    end
    return true, token, perm
end

function utils.check_permission(...)
    local ok, token, perm = utils.check_auth()
    if not ok then
        return utils.error_exit(ngx.HTTP_UNAUTHORIZED, "unauthorized")
    end

    if not perm then
        return utils.error_exit(ngx.HTTP_FORBIDDEN, "forbidden")
    end

    local perm_list = split(perm, ",")
    for _, item in ipairs { ... } do
        if (in_array(perm_list, item)) then
            return true, token
        end
    end
    utils.error_exit(ngx.HTTP_FORBIDDEN, "forbidden")
end

function utils.error_exit(status, msg)
    local res = {}
    res["Code"] = status
    res["Message"] = msg
    ngx.status = status
    ngx.header['Content-Type'] = 'application/json; charset=utf-8'
    ngx.say(json.encode(res))
    ngx.exit(ngx.HTTP_OK)
end


function utils.get_item_index(total, page_size, current_page)
    if (total <= 0 or page_size <= 0 or current_page <= 0) then
        return 0, -1
    end

    local min = page_size * (current_page - 1) + 1
    local max = page_size * current_page
    if (total < max) then
        max = total
    end

    return min, max
end


function utils.call_shell(cmd)
    local ok, retcode, stdout, stderr = pl_utils.executeex(cmd)
    return ok, retcode, stdout, stderr
end

local function dbus_send(bus, wkn, obj, inf, method, ...)
    local failure = function(cause)
        utils.log_error(cause)
        return false, nil, cause
    end

    local conn, err_c = ldbus.connection.open(bus)
    if not conn then
        return failure("failed to open connection: " .. tostring(err_c))
    end

    local reg, err_r = ldbus.bus.register(conn)
    if not reg then
        return failure("failed to register connection: " .. tostring(err_r))
    end

    local msg = ldbus.message.new_method_call(wkn, obj, inf, method)
    if not msg then
        local err_m = string.format("failed to create method call for (%s, %s, %s.%s)",
                wkn, obj, inf, method)
        return failure(err_m)
    end

    local iter = ldbus.message.iter.new()
    msg:iter_init_append(iter)

    for _, arg in ipairs { ... } do
        iter:append_basic(arg)
    end

    local timeout = 20
    if (wkn == const.function_compute.wkn) then timeout = 1800 end

    local reply, err_r = conn:send_with_reply_and_block(msg, timeout)
    if not reply then
        local err_f = string.format("failed to send method call for %s: %s", method, err_r)
        return failure(err_f)
    end

    if not reply:iter_init(iter) then
        return failure("reply message has no arguments")
    end

    local ret = {}
    while iter:get_arg_type() ~= nil do
        table.insert(ret, iter:get_basic())
        iter:next()
    end
    return true, ret, nil
end


function utils.dbus_call_edge(wkn, obj, inf, method, ...)
    return dbus_send("unix:path=/tmp/var/run/mbusd/mbusd_socket", wkn, obj, inf, method, ...)
end


function utils.set_data(key, val)
    local ok, ret, err = utils.dbus_call_edge(
            "iot.dmp.configmanager",
            "/iot/dmp/configmanager",
            "iot.dmp.configmanager",
            "set_config", key, val)

    if not ok then
        return false, err
    end

    local code = tonumber(ret[1])
    if code == 0 then
        return true, nil
    end

    return false, "error: no. = " .. code
end

function utils.get_standard_tsl_key_list()
    local ok, ret, stderr = utils.dbus_call_edge(
            "iot.dmp.configmanager",
            "/iot/dmp/configmanager",
            "iot.dmp.configmanager",
            "get_config_keylist",
            const.data_key.tsl.standard_tsl_prefix .. "*")
    if (ok) then
        local int32_code = tonumber(ret[1])
        if (0 == int32_code) then
            msg = ret[2]
            local ret = {}
            for k, v in pairs(json.decode(msg)) do
                if string.sub(v, 1, #const.data_key.tsl.extend_tsl_prefix) ~= const.data_key.tsl.extend_tsl_prefix then
                    table.insert(ret, v)
                end
            end

            if ("" ~= msg) then
                return json.encode(ret)
            end
        end
        -- ngx.log(ngx.WARN, "int code:"..int32_code.."value of the key:", key, " is empty. ", stdout[2])
        return "{}"
    end

    utils.log_error("get data failed" .. ret[2] .. stderr)
    utils.error_exit(500, "get data failed")
end

function utils.get_extend_tsl_key_list()
    local ok, stdout, stderr = utils.dbus_call_edge("iot.dmp.configmanager",
                                "/iot/dmp/configmanager",
                                "iot.dmp.configmanager",
                                "get_config_keylist",
                                const.data_key.tsl.extend_tsl_prefix.."*")
    if (ok) then
        int32_code = tonumber(stdout[1])
        if (0 == int32_code) then
            msg = stdout[2]
            if ("" ~= msg) then
                return msg
            end
        end
        -- ngx.log(ngx.WARN, "int code:"..int32_code.."value of the key:", key, " is empty. ", stdout[2])
        return "{}"
    end

    utils.log_error("get data failed" .. stdout[2] .. stderr)
    utils.error_exit(500, "get data failed")
end

function utils.get_data(key)
    local ok, ret, err = utils.dbus_call_edge(
            "iot.dmp.configmanager",
            "/iot/dmp/configmanager",
            "iot.dmp.configmanager",
            "get_config",
            key)
    if not ok then
        return false, "get config failed: " .. err
    end

    local code = tonumber(ret[1])
    if code == 100100 then
        return false, "key does not exit: " .. key
    end
    if code ~= 0 then
        return false, "get config failed: " .. code
    end
    return true, ret[2]
end


local function is_match(orign, rule, len_min, len_max)
    if (nil == orign) then
        return false
    end

    if (string.match(orign, rule)) then
        len = string.len(orign)
        if (len >= len_min and len <= len_max) then
            return true
        end
    end

    return false
end


local function check_invalid_exit(orign, rule, len_min, len_max, msg)
    if (not is_match(orign, rule, len_min, len_max)) then
        utils.error_exit(400, msg)
    end

    return true
end


function utils.check_username(username)
    if type(username) ~= "string" or #username == 0 then
        utils.error_exit(400, "username not a string or is empty")
    end
    return check_invalid_exit(username, "^%w+$", 5, 20, "invalid username")
end


function utils.check_password(password)
    if type(password) ~= "string" or #password == 0 then
        utils.error_exit(400, "password not a string or is empty")
    end

    if (64 == string.len(password)) then --password, 64 chars
        return true
    end
    return check_invalid_exit(password, "^%w+$", 5, 20, "invalid password")
end


function utils.check_pk(pk)
    return check_invalid_exit(pk, "^%w+$", 11, 11, "invalid product key")
end


function utils.check_dn(dn)
    -- 支持英文字母、数字和特殊字符-_@.:，长度限制4~32
    return check_invalid_exit(dn, "^[%w%_%-%.%@%:]+$", 4, 32, "invalid device name")
end


function utils.check_ds(ds)
    return check_invalid_exit(ds, "^%w+$", 32, 32, "invalid device secret")
end


function utils.check_authmode(authmode)
    if ("Auto" == authmode or "Manual" == authmode) then
        return true
    end

    utils.error_exit(400, "invalid authmode")
end


function utils.check_tag(tag)
    -- empty or table list
    type_tag = type(tag)
    if (nil == tag or "string" == type_tag or "table" == type_tag) then
        return true
    end

    utils.error_exit(400, "invalid tag")
end


function utils.check_permission_format(permission)
    valid = {"GatewayManage", "DeviceManage", "DriverManage", "LogManage", "UserManage", "MessageRoute", "TSLManage", "StreamCompute", "FunctionComputeManage"}

    for _, val in ipairs(permission) do
        if (not in_array(valid, val)) then
            utils.error_exit(400, "invalid permission format")
        end
    end

    return true
end


function utils.check_current_page( cur_page )
    if (cur_page and "number" == type(cur_page) and cur_page > 0) then
        return true
    end

    utils.error_exit(400, "invalid current page")
end

function utils.check_page_size_allow_zero( page_size )
    
    if page_size == nil then
        return 0
    end

    if type(page_size) == "number" and page_size >= 0 then
        return  page_size
    end

    if type(page_size) == "string" and #page_size == 0 then
        return 0
    end

    utils.error_exit(400, "invalid page size")
end


function utils.check_page_size( page_size )
    if (page_size and "number" == type(page_size) and page_size > 0) then
        return true
    end

    utils.error_exit(400, "invalid page size")
end


-- get caller infomation from debug.traceback
-- output like: /src/utils.lua:370: in function 'log_info':
local function get_log_prefix()

    --[[ sample of traceback
    stack traceback:
	    ./src/utils.lua:350: in function 'get_log_prefix'
	    ./src/utils.lua:370: in function 'log_info'
	    /var/nginx/eweb/src/login.lua:52: in function </var/nginx/eweb/src/login.lua:1>, client: 30.6.57.149,
    ]]--

    tb = debug.traceback()
    tb_list = split(tb, '\n')
    assert(tb_list[4], "tb_list[4] should never be nil")

    log_prefix = string.gsub(tb_list[4], "^(%s*)(.+)$", "%2")
    -- log_prefix = string.gsub(tb_list[4], "^(%s+)(.+) in function '(%w+)'", "%2%3(): ")
    -- if (nil == log_prefix) then
    --     log_prefix = string.gsub(tb[4], "^(%s+)(.+) in function <(.+)>", "%2%3(): ")
    -- end
    --return log_prefix

    return log_prefix
end

local function check_init_log()
    local memkv = ngx.shared.memkv
    local flag = memkv:get("flag_log_inited") 
    
    if not flag then
        logger.log_init("web_console", 1, 0, 2)
        ngx.ctx.init_logger = 1
        memkv:set("flag_log_inited", 1)
        ngx.log(ngx.ERR, "init logger")       
    end
end

local function log_dump(lvl, msg)
    local g_logger_switch = 1
    if (g_logger_switch == 1) then
        local func_dispatch = nil
        check_init_log()
        if (lvl == ngx.DEBUG) then
            func_dispatch = logger.log_d
        elseif (lvl == ngx.INFO) then
            func_dispatch = logger.log_i
        elseif (lvl == ngx.ERR) then
            func_dispatch = logger.log_e
        elseif (lvl == ngx.CRIT) then
            func_dispatch = logger.log_m
        elseif (lvl == ngx.WARN) then
            func_dispatch = logger.log_w
        elseif (lvl == ngx.INFO) then
            func_dispatch = logger.log_i
        end
        func_dispatch("local_console", msg)
    else
        ngx.log(lvl, msg)
    end
end

function utils.log_debug(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)

    log_dump(ngx.DEBUG, msg)
end

function utils.log_info(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)
    log_dump(ngx.INFO, msg)
end

function utils.log_warn(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)
    log_dump(ngx.WARN, msg)
end

function utils.log_error(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)
    log_dump(ngx.ERR, msg)
end

function utils.log_critical(fmt, ... )
    fmt = get_log_prefix() .. fmt
    msg = string.format(fmt, ...)
    log_dump(ngx.CRIT, msg)
end


function utils.set_gw_config(pk, dn, ds, auth_mode)
    local ok1, msg1 = utils.set_data(const.data_key.gwconfig.productkey, pk)
    local ok2, msg2 = utils.set_data(const.data_key.gwconfig.devicename, dn)
    local ok3, msg3 = utils.set_data(const.data_key.gwconfig.devicesecret, ds)
    if (not (ok1 and ok2 and ok3)) then
        utils.log_error("set gateway config failed" .. msg1 .. msg2 .. msg3)
        utils.error_exit(500, "set data failed")
    end

    local auto_auth = "off" -- in off mode by default
    if ("Auto" == auth_mode) then
        auto_auth = "on"
    end

    utils.set_data(const.data_key.dimu.auto_auth_sw, auto_auth)
    if (not ok) then
        utils.log_error("set gateway auto auth value failed" .. msg)
        utils.error_exit(500, "set data failed")
    end
end


-- return: pk, dn, secret, authmode, tags
function utils.get_gw_config( )
    -- get gateway 3tuple
    local ok1, msg_pk = utils.get_data(const.data_key.gwconfig.productkey)
    local ok2, msg_dn = utils.get_data(const.data_key.gwconfig.devicename)
    local ok3, msg_ds = utils.get_data(const.data_key.gwconfig.devicesecret)
    if (not (ok1 and ok2 and ok3)) then
        utils.log_error("get gateway config failed" .. msg_pk .. msg_dn .. msg_ds)
        utils.error_exit(500, "get data failed")
    end

    if (nil == msg_pk or nil == msg_dn or nil == msg_ds) then
        return nil, nil, nil, nil
    end

    -- get auth mode
    local auth_mode = "Manual" -- 'Manual' mode by default
    local auto_auth = utils.get_data(const.data_key.dimu.auto_auth_sw)
    if (not ok) then
        utils.log_error("get gateway auto auth value failed" .. msg)
        utils.error_exit(500, "get data failed")
    end
    if (auto_auth and "on" == auto_auth) then
        auth_mode = "Auto"
    end

    return msg_pk, msg_dn, msg_ds, auth_mode
end


function utils.set_driver_config(driver, config)
    local key = const.data_key.driver.prefix .. driver
    local ok, msg = utils.set_data(key, config)
    if (not ok) then
        utils.log_error("store driver config failed" .. msg)
        utils.error_exit(500, "store config failed")
    end
end

function utils.get_driver_config(driver) 
    local key = const.data_key.driver.prefix .. driver
    local ok, msg = utils.get_data(key)
    if (not ok) then
        -- utils.error_exit(500, "get config failed")
        return nil
    end

    return msg
end

function utils.set_device_tags(pk, dn, tags)
    local cjson = require('cjson')
    local device_tags = {}
    local params = {}
    params["params"] = {}
    params["params"]["devices"] = {pk .. '_' .. dn}
    params["params"]["tags"] = {}
    for _, item in ipairs(tags) do
        local res = utils.split_str(item, '=')
        params["params"]["tags"][res[1]] = res[2]
    end

    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                    const.dimu.obj,
                                    const.dimu.inf,
                                    "setDeviceTags",
                                    cjson.encode(params))

    if (ok == "ok") then
        local ok, msg = pcall(json.decode, msg_raw[1])
        if ok then
            if 0 == msg["code"] then
                utils.error_exit(200, json.encode(""))
            else
                local ret = {}
                local ret_error = {}
                if not (msg["params"]["failed_message"] and type(msg["params"]["failed_message"]) == "table" )then
                 utils.error_exit(400, "failed to set device tags.")
                end
                for k, v in pairs(msg["params"]["failed_message"]) do
                    if v["code"] ~= 0 then
                        local item = {}
                        local tmp = utils.split_str(v["params"]["device"], "_")
                        item["ProductKey"] = tmp[1]
                        item["DeviceName"] = table.concat(tmp, "_", 2)
                        item["FaileMessage"] = v["message"]
                        table.insert(ret_error, item)
                    end
                end
                ret["code"] = 400
                ret["data"] = ret_error
                ret["message"] = ""
                ngx.say(json.encode(ret))
                ngx.exit(ngx.HTTP_OK)
            end
        end
        if (ok and 0 == msg["code"]) then
            utils.error_exit(200, json.encode(""))
        end
    else
        utils.log_error("setDeviceTags failed" .. msg)
        utils.error_exit(500, "store device tags failed")
    end
end

function utils.get_device_tags(pk, dn)
    local device_tags = {}
    
    if(pk == nil or dn == nil) then
        return device_tags
    end

    local cloudid = pk .. '_' .. dn
    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                    const.dimu.obj,
                                    const.dimu.inf,
                                    "getDeviceTags",
                                    '{"params": ["' .. cloudid .. '"]}')
    if (ok and msg_raw) then
        local ok, msg = pcall(json.decode, msg_raw[1])
        if (ok and 0 == msg["code"]) then
            local tags = msg["params"][cloudid]
            for key, val in pairs(tags) do
                table.insert(device_tags, key .. "=" .. val)
            end
        end
    else
        utils.log_error("get nothing from getDeviceTags:" .. err)
    end

    return device_tags
end

function utils.sync_device_tags()
    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                    const.dimu.obj,
                                    const.dimu.inf,
                                    "syncTag")

    if (ok) and (tonumber(msg_raw[1]) == 0) then
        return true, "ok"
    else
        utils.log_error("syncDeviceTags failed : " .. msg_raw[1])
        return false, msg_raw[1]
    end
end

function utils.set_product_key_list( product_keys )
    local ok, msg = utils.set_data(const.data_key.eweb.product_key_list, json.encode(product_keys))
    if (ok) then
        return true
    end

    utils.log_error("set product key list failed" .. msg)
    utils.error_exit(500, "store product key list failed")
end


function utils.get_product_key_list( )
    local ok, msg = utils.get_data(const.data_key.eweb.product_key_list)
    if (ok and msg) then
        local ok, list = pcall(json.decode, msg) 
        if (ok) then
            return list
        end
    else 
        return {}
    end

    utils.log_error("get device tuple data failed" .. msg)
    utils.error_exit(500, "get device tuple failed")
end

function utils.get_tsl(product)
    return utils.get_data(const.data_key.tsl.standard_tsl_prefix .. product)
end

return utils

------ EOF ------
